import ninfo
from tests.common import Wrapper

def test_plugin_loading():
    test_plugins = {
        "a": Wrapper("tests.plug_a"),
    }
    n=ninfo.Ninfo(plugin_modules=test_plugins)
    assert 'a' in n
    assert 'b' not in n

    plugin = n.get_plugin('a')
    assert plugin is not None

    plugin = n.get_plugin('b')
    assert plugin is None

def test_plugin_lazy_loading():
    test_plugins = {
        "a": Wrapper("tests.plug_a"),
    }
    n=ninfo.Ninfo(plugin_modules=test_plugins)
    assert 'a' in n
    assert 'a' not in n.plugin_instances

    plugin = n.get_plugin('a')
    assert plugin is not None
    assert 'a' in n.plugin_instances

def test_plugin_lazy_init():
    test_plugins = {
        "a": Wrapper("tests.plug_a"),
    }
    n=ninfo.Ninfo(plugin_modules=test_plugins)
    plugin = n.get_plugin('a')
    assert plugin is not None
    assert 'a' in n.plugin_instances

    assert plugin.initialized is False

    res = n.get_info("a", "example.com")
    assert plugin.initialized is True

    assert res == "AAAAAAAAAAA"

def test_plugin_compatible_types():
    test_plugins = {
        "a": Wrapper("tests.plug_a"),
    }
    n=ninfo.Ninfo(plugin_modules=test_plugins)

    cases = [
        ("example.com", True),
        ("1.2.3.4", True),
        ("00:11:22:33:44:55", False),
    ]
    for arg, expected in cases:
        yield _plugin_compatible_type_case, n, arg, expected

def _plugin_compatible_type_case(n, arg, expected):
    assert n.compatible_argument("a", arg) == expected



import ninfo
from tests.common import Wrapper

def test_plugin_cloning_no_config():
    test_plugins = {
        "a": Wrapper("tests.plug_a"),
    }
    n=ninfo.Ninfo(plugin_modules=test_plugins)
    assert 'a' in n
    assert 'b' not in n

def test_plugin_cloning_with_config():
    test_plugins = {
        "a": Wrapper("tests.plug_a"),
    }
    n=ninfo.Ninfo(plugin_modules=test_plugins, config_file="tests/clone.cfg")
    assert 'a' in n
    assert 'b' in n

def test_plugin_cloning_cloned_config():
    test_plugins = {
        "a": Wrapper("tests.plug_a"),
    }
    n=ninfo.Ninfo(plugin_modules=test_plugins, config_file="tests/clone.cfg")
    a = n.get_plugin('a')
    b = n.get_plugin('b')

    assert a.plugin_config == {"one": '1', "two": '2'}
    assert b.plugin_config == {"clone": 'a', "one": 'one', "two": '2', "three": '3'}



import ieeemac
import IPy
import re


def isip(arg):
    """is arg an ip address?"""
    try:
        x = IPy.IP(arg)
        return x.version()
    except ValueError:
        return False


_hash_re = re.compile("^[0-9a-fA-F]+$")


def ishash(arg):
    return bool(_hash_re.match(arg))


def get_type(arg):
    """Return a list of possible types for the type of the
    argument (mac, ip, hostname, url,hostport, and/or username)"""

    potential_types = []

    ip_types = {
        (4, False): "ip",
        (6, False): "ip6",
        (4, True): "cidr",
        (6, True): "cidr6",
    }

    if ieeemac.ismac(arg):
        potential_types.append("mac")
    # macs can look like hashes, so check them first
    if ishash(arg):
        potential_types.append("hash")
    ipver = isip(arg)
    if ipver:
        potential_types.append(ip_types[(ipver, "/" in arg)])

    parts = arg.split(":")
    if "://" in arg:
        potential_types.append("url")
    elif len(parts) == 2 and parts[1].isdigit():
        potential_types.append("hostport")
    elif "." in arg and not ipver:
        potential_types.append("hostname")

    potential_types.append("username")

    return potential_types


def is_local(networks, ip):
    """Return True if `ip` is in `networks`"""
    for n in networks:
        if ip in n:
            return True
    return False


def parse_query(s):
    """Split the query up into arguments and options
    This should be easy to do using a regex, but I can't figure it out.
    so, state machine it is."""

    args = []
    options = {}

    # regex too complicated, state machine easy
    # first split the string into components, respecting quotes
    parts = []
    in_quote = False
    p = ""
    for l in s:
        if l.isspace() and not in_quote:
            parts.append(p)
            p = ""
        elif l in "\"'" and not in_quote:
            in_quote = l
        elif l == in_quote:
            in_quote = False
        else:
            p += l
    if p:
        parts.append(p)

    # now separate each argument or option
    for arg in parts:
        if "=" in arg and "//" not in arg:  # has a = but doesn't look like a URL
            k, v = arg.split("=", 1)
            v = v.strip('"')  # remove the quotes
            options[k] = v
        else:
            args.append(arg)
    return args, options


def fmt_dict_array(ar, order=None, header=True):
    # allow for an empty array
    # if the array is empty and no order is paseed, return ""
    # otherwise generate just the header
    if not ar:
        if not order:
            return ""
        first = dict([(x, x) for x in order])
    else:
        first = ar[0]
    if header:  # insert a new row of just the keys, the next loop will do the rest
        new = dict([(x, str(x).capitalize()) for x in first.keys()])
        new = [new]
    else:
        new = []
    # figure out what the longest column in each row is
    maxes = dict([(a, len(str(b))) for a, b in first.items()])
    for x in new + ar:
        for k, v in x.items():
            size = len(str(v))
            if size > maxes[k]:
                maxes[k] = size

    maxesfmt = dict([(x, "%%-%ds" % (y + 1)) for x, y in maxes.items()])
    if order:
        it = order
    else:
        it = first  # will iterate over the keys however

    table = []

    for x in new + ar:
        row = []
        for c in it:
            row.append(maxesfmt[c] % x[c])
        line = " ".join(row).strip()
        table.append(line)

    return "\n".join(table)


def unique(lst):
    out = []
    seen = set()
    for x in lst:
        if x not in seen:
            out.append(x)
            seen.add(x)
    return out



"""Requires the splunk-sdk and the following in the ninfo config:

[splunk]
host = splunk-backend.example.com
web = splunk-head.example.com
port = 8089
username = infosearch
password = password_for_info_to_use
"""

from ninfo import PluginBase

import logging

logger = logging.getLogger(__name__)


class SplunkBase(PluginBase):
    """This plugin is a base that can be used for splunk searches

    subclass it and override 'TEMPLATE' or get_info
    """

    TEMPLATE = None

    def setup(self):
        import splunklib.client as client
        import splunklib.results as results

        self.client = client
        self.splunklibresults = results

    def connect(self):
        sc = self.config["splunk"]
        host = sc["host"]
        port = sc["port"]
        username = sc["username"]
        password = sc["password"]
        self.s = self.client.connect(
            host=host, port=port, username=username, password=password
        )

    def do_search(self, query):
        self.connect()
        logger.debug("Searching splunk for '%s'" % query)
        q = "search " + query
        rr = self.splunklibresults.ResultsReader(self.s.jobs.export(q))
        events = []
        for result in rr:
            if isinstance(result, self.splunklibresults.Message):
                # Diagnostic messages may be returned in the results
                logger.debug("%s: %s" % (result.type, result.message))
            elif isinstance(result, dict):
                # Normal events are returned as dicts
                events.append(result)
        return events

    def get_info(self, arg):
        if not self.TEMPLATE:
            raise NotImplementedError(
                "You must override self.TEMPLATE or self.get_info"
            )

        query = self.TEMPLATE % arg
        events = self.do_search(query)
        return {"events": events}



from ninfo import util

def test_is_ip():
    cases = (
        ("1.2.3.4", 4),
        ("123.123.123.123", 4),
        ("123.123.123.321", False),
        ("1.2.3.4a", False),
        ("a1.2.3.4", False),
        ("blah", False),
    )

    for ip, valid in cases:
        yield ip_case, ip, valid

def ip_case(ip, valid):
    assert util.isip(ip) == valid

def test_get_type():
    cases = (
        ("1.2.3.4", "ip"),
        ("2001:4860:800e::69", "ip6"),
        ("001122334455", "mac"),
        ("blah.com", "hostname"),
        ("blah", "username"),
        ("1.2.3.0/24", "cidr"),
        ("2001:4860:800e::0/48", "cidr6"),
        ("2eadc4e74f9d03f3bd8e9865000d282ecb96f108", "hash"),
        ("69fe264e8ff51121841b0896908c419923c0e31352100ca55a0740d6a00fe9a4", "hash"),
        ("127.0.0.1:80", "hostport"),
        ("https://192.168.2.1:443", "url"),
    )
    for x, t in cases:
        yield type_case, x, t

def type_case(x, expected_type):
    assert expected_type in util.get_type(x)


import IPy
def test_is_local():
    networks = ["192.168.1.0/24","192.168.99.0/24"]
    networks = [IPy.IP(n) for n in networks]

    cases = (
        ("1.2.3.4", False),
        ("192.168.1.33", True),
        ("192.168.99.33", True),
        ("192.168.100.33", False),
    )

    for ip, result in cases:
        yield is_local_case, networks, ip, result

def is_local_case(networks, ip, result):
    assert util.is_local(networks, ip) == result


def test_query_parsing():
    cases = (
        ('one two', (['one', 'two'], {})),
        ('arg key=value', (['arg'], {'key': 'value'})),
        ('one two key=value', (['one','two'], {'key': 'value'})),
        ('one two key=value b=c', (['one','two'], {'key': 'value', 'b': 'c'})),
        ('arg key="spaced value"', (['arg'], {'key': 'spaced value'})),
        ('arg two key="spaced value" b="c d"', (['arg','two'], {'key': 'spaced value', 'b': "c d"})),
        ('1.2.3.4', (['1.2.3.4'], {})),
        ('1.2.3.4 time="2012-04-19 11:50"', (['1.2.3.4'], {'time':'2012-04-19 11:50'})),
        ('1.2.3.4 time="2012-04-19 11:50:22"', (['1.2.3.4'], {'time':'2012-04-19 11:50:22'})),
        ('00:11:22:33:44:55', (['00:11:22:33:44:55'], {})),
        ('http://example.com', (['http://example.com'], {})),
    )

    for input, output in cases:
        yield query_parse_case, input, output

def query_parse_case(input, output):
    assert util.parse_query(input) == output, "%r != %r" % (util.parse_query(input), output)

def test_unique():
    cases = (
        (['a','a'], ['a']),
        (['a','b','a'], ['a','b']),
    )
    for input, output in cases:
        yield unique_case, input, output

def unique_case(input, output):
    assert util.unique(input) == output



from pkg_resources import iter_entry_points

__version__ = "1.0.1"

import memcache
import logging

logger = logging.getLogger("ninfo")

import os

try:
    import ConfigParser
except ImportError:
    import configparser as ConfigParser

import json

from mako.template import Template

try:
    from inspect import getargspec
except ImportError:
    from inspect import getfullargspec as getargspec

from inspect import getsourcefile

from ninfo import util
import IPy


def clean_cache_key(s):
    return "".join(c for c in s if 32 < ord(c) < 127).encode("utf-8")


class PluginError(Exception):
    def __init__(self, message, cause):
        super(PluginError, self).__init__(message + ", caused by " + repr(cause))
        self.cause = cause


class PluginInitError(PluginError):
    pass


class PluginBase(object):
    cache_timeout = 60 * 60
    local = True
    remote = True
    template = True
    options = {}
    converters = {}

    def __init__(self, config=None, plugin_config=None):
        self.long_desription = self.__doc__
        if config is None:
            config = {}
        if plugin_config is None:
            plugin_config = {}
        self.config = config
        self.plugin_config = plugin_config
        self.initialized = False
        self._name = self.name
        self._template = self.name

        if "template_override" in plugin_config:
            self._template = plugin_config["template_override"]

        if "disabled" in plugin_config:
            return

    def init(self):
        if self.initialized:
            return True
        try:
            self.initialized = self.setup() != False
            return self.initialized
        except Exception as e:
            logger.exception("Error initializing plugin %s" % self.name)
            raise PluginInitError("Error initializing plugin %s" % self.name, cause=e)

    def as_json(self):
        return {
            "name": self.name,
            "title": self.title,
            "description": self.description,
            "long_description": self.long_description,
            "cache_timeout": self.cache_timeout,
            "types": self.types,
            "local": self.local,
            "remote": self.remote,
        }

    def setup(self):
        return True

    def to_json(self, result):
        return result

    def get_info_json(self, arg):
        return self.to_json(self.get_info(arg))

    def get_info_text(self, arg):
        result = self.get_info(arg)
        return self.render_template("text", arg, result)

    def get_info_html(self, arg):
        result = self.get_info(arg)
        return self.render_template("html", arg, result)

    def _do_render(self, filename, arg, result):
        if filename is None:
            return str(result)
        kw = {}
        if "html" in filename:
            kw["default_filters"] = ["h"]
        t = Template(filename=filename, **kw)
        out = t.render(
            arg=arg,
            plugin=self,
            config=self.config,
            plugin_config=self.plugin_config,
            **result
        )
        return out.lstrip()

    def render_template(self, output_type, arg, result):
        if not result:
            return ""
        filename = self.get_template(output_type)
        if filename is None and output_type == "html":
            filename = self.get_template("text")
            out = self._do_render(filename, arg, result)
            if out:
                out = "<pre>" + out + "</pre>"
            return out

        out = self._do_render(filename, arg, result)
        return out

    def get_template(self, output_type):
        code = getsourcefile(self.__class__)
        path = os.path.dirname(code)
        filename = "%s_template_%s.mako" % (self._template, output_type)
        template = os.path.join(path, filename)
        if os.path.exists(template):
            return template

    def get_converter(self, a, b):
        func = self.converters.get((a, b))
        if func:
            return getattr(self, func)


class Ninfo:
    def __init__(self, config_file=None, plugin_modules=None):
        self.plugin_instances = {}
        if plugin_modules:
            self.plugin_modules = plugin_modules
        else:
            self.plugin_modules = {}
            for ep in iter_entry_points(group="ninfo.plugin"):
                self.plugin_modules[ep.name] = ep

        self.read_config(config_file)

    def __contains__(self, plugin):
        return plugin in self.plugin_modules

    def read_config(self, config_file):
        cp = ConfigParser.ConfigParser()
        if config_file:
            cp.read([config_file])
        elif os.getenv("INFO_CONFIG_FILE"):
            cp.read([os.getenv("INFO_CONFIG_FILE")])
        else:
            cp.read(
                ["/etc/ninfo.conf", os.path.expanduser("~/.ninfo.ini"), "ninfo.ini"]
            )
            # cp.read(["ninfo.ini",os.path.expanduser("~/.ninfo.ini"),"/etc/ninfo.ini"])
        # return a simple nested dictionary structure from the config
        self.config = dict((s, dict(cp.items(s))) for s in cp.sections())

        # Remove disabled plugins
        for section in self.config:
            if "disabled" in self.config[section]:
                plugin_name = section.split(":")[1]
                try:
                    del self.plugin_modules[plugin_name]
                except KeyError:
                    logger.debug(
                        "Plugin %s is disabled in .ini file, but was not found."
                        % plugin_name
                    )
                else:
                    logger.info("Plugin %s disabled via .ini file." % plugin_name)

        # Add entries for a cloned plugin configuration
        for section in self.config:
            clone = self.config[section].get("clone")
            disabled = self.config[section].get("disabled")
            if clone and not disabled and clone in self.plugin_modules:
                plugin_name = section.split(":")[1]
                self.plugin_modules[plugin_name] = self.plugin_modules[clone]

        if "general" not in self.config:
            self.cache = None
            self.local_networks = []
            return

        self.cache_host = self.config["general"].get("memcache_host")
        if self.cache_host:
            self.cache = memcache.Client([self.cache_host])
        else:
            self.cache = None

        networks_str = self.config["general"].get("local_networks", "")
        self.local_networks = [IPy.IP(n) for n in networks_str.split(",")]

    def is_local(self, arg):
        for t in util.get_type(arg):
            if t in ["ip", "ip6"]:
                return util.is_local(self.local_networks, arg)
        return False

    def compatible_argument(self, plugin, arg):
        plug = self.get_plugin(plugin)
        if not plug:
            logger.debug("Skipping plugin %s because plugin does not exist" % plugin)
            return False
        potential_arg_types = util.get_type(arg)
        for arg_type in potential_arg_types:
            if arg_type in plug.types:
                if arg_type in ["ip", "ip6"]:
                    if plug.local == False and self.is_local(arg):
                        logger.debug("Skipping plugin %s because arg is local" % plugin)
                        return False
                    if plug.remote == False and not self.is_local(arg):
                        logger.debug(
                            "Skipping plugin %s because arg is remote" % plugin
                        )
                        return False
                return True
        logger.debug("Skipping plugin %s because arg is the wrong type" % plugin)
        return False

    def get_plugin(self, plugin):
        if plugin not in self.plugin_modules:
            return None
        if plugin in self.plugin_instances:
            return self.plugin_instances[plugin]

        plugin_config_key = "plugin:" + plugin
        plugin_config = self.config.get(plugin_config_key, {})
        if "disabled" in plugin_config:
            return None

        # If this plugin was cloned, merge its config on top of the config from
        # the cloned plugin
        if "clone" in plugin_config:
            clone_key = "plugin:" + plugin_config["clone"]
            new_cfg = self.config.get(clone_key, {}).copy()
            new_cfg.update(plugin_config)
            plugin_config = new_cfg

        try:
            cls = self.plugin_modules[plugin].load().plugin_class
            cls.long_description = cls.__doc__
        except:
            logger.exception("Error loading plugin %s" % plugin)
            if plugin in self.plugin_modules:
                del self.plugin_modules[plugin]
            return

        instance = cls(config=self.config, plugin_config=plugin_config)
        # override any plugin metadata for the case of cloned plugins(or otherwise)
        instance.name = plugin
        for field in "name", "title", "description":
            if field in plugin_config:
                setattr(instance, field, plugin_config[field])
        self.plugin_instances[plugin] = instance
        return instance

    @property
    def plugins(self):
        plugins = [self.get_plugin(p) for p in sorted(self.plugin_modules.keys())]
        return [p for p in plugins if p]

    def get_info(self, plugin, arg, options={}, retries=1):
        """Call `plugin` with `arg`, `options` and cache and return the result"""
        if not self.compatible_argument(plugin, arg):
            return None
        plugin_obj = self.get_plugin(plugin)
        timeout = plugin_obj.cache_timeout
        if self.cache and timeout:
            KEY = "ninfo:%s:%s" % (plugin, arg)
            KEY += ":" + ":".join(
                [
                    "%s=%s" % (key, value)
                    for (key, value) in options.items()
                    if key in plugin_obj.options
                ]
            )
            # Remove non-alphanumerics from key, to be safe
            KEY = clean_cache_key(KEY)
            ret = self.cache.get(KEY)
            if ret:
                return ret[1]

        try:
            plugin_obj.init()
            get_info_args = len(getargspec(plugin_obj.get_info)[0])
            if get_info_args == 3:
                # This plugin supports context.
                ret = plugin_obj.get_info(arg, options)
            else:
                # This plugin doesn't.
                ret = plugin_obj.get_info(arg)
            if self.cache and timeout:
                self.cache.set(KEY, (True, ret), timeout)
            return ret
        except PluginInitError:
            raise
        except Exception as e:
            logger.exception("Error running plugin %s" % plugin)
            if retries:
                return self.get_info(plugin, arg, options, retries - 1)
            raise PluginError("Error running plugin %s" % plugin, cause=e)

    def get_info_json(self, plugin, arg, options={}):
        result = self.get_info(plugin, arg, options)

        p = self.get_plugin(plugin)
        return p.to_json(result)

    def get_info_text(self, plugin, arg, options={}):
        result = self.get_info(plugin, arg, options)

        p = self.get_plugin(plugin)
        return p.render_template("text", arg, result)

    def get_info_html(self, plugin, arg, options={}):
        result = self.get_info(plugin, arg, options)

        p = self.get_plugin(plugin)
        return p.render_template("html", arg, result)

    def get_info_iter(self, arg, plugins=None, options={}):
        for p in sorted(self.plugins, key=lambda p: p.name):
            if plugins and p.name not in plugins:
                continue
            if not self.compatible_argument(p.name, arg):
                continue
            try:
                result = self.get_info(p.name, arg, options)
                yield p, result
            except PluginError:
                logger.exception("Error running plugin %s", p.name)

    def get_info_dict(self, arg, plugins=None, options={}):
        res = {}
        for p, result in self.get_info_iter(arg, plugins, options):
            res[p.name] = result
        return res

    def show_info(self, arg, plugins=None, options={}):
        for p, result in self.get_info_iter(arg, plugins, options):
            print("*** %s (%s) ***" % (p.title, p.description))
            print(p.render_template("text", arg, result))

    def convert(self, arg, to_type):
        potential_arg_types = util.get_type(arg)
        for p in self.plugins:
            for arg_type in potential_arg_types:
                try:
                    if (arg_type, to_type) in p.converters:
                        p.init()
                        yield p.name, p.get_converter(arg_type, to_type)(arg)
                except Exception:
                    logger.exception("Error running plugin %s", p.name)


def main():
    logging.basicConfig(level=logging.INFO)
    # requests logs stuff at level INFO
    logging.getLogger("requests.packages.urllib3").setLevel(logging.ERROR)

    from optparse import OptionParser

    parser = OptionParser(usage="usage: %prog [options] [addresses]")
    parser.add_option(
        "-p",
        "--plugin",
        dest="plugins",
        action="append",
        help="The plugin to run",
        default=None,
    )
    parser.add_option(
        "-j",
        "--json",
        dest="json",
        action="store_true",
        help="Output json instead of rendering",
        default=False,
    )
    parser.add_option("-l", "--list", dest="list", action="store_true", default=False)
    (options, complete_args) = parser.parse_args()

    p = Ninfo()
    if options.list:
        print("%-20s %-20s %s" % ("Name", "Title", "Description"))
        for pl in p.plugins:
            print("%-20s %-20s %s" % (pl.name, pl.title, pl.description))
        return

    context_options = {}
    args = []
    for arg in complete_args:
        if "=" in arg:
            (ctxt_name, ctxt_value) = arg.split("=", 1)
            context_options[ctxt_name] = ctxt_value
        else:
            args.append(arg)

    plugins = options.plugins or None
    json_output = []
    for arg in args:
        if len(args) != 1 and not options.json:
            print("=== %s === " % (arg,))
        if options.json:
            json_output.append({"arg": arg, "data": p.get_info_dict(arg, plugins=plugins, options=context_options)})
        else:
            p.show_info(arg, plugins=plugins, options=context_options)

    if options.json:
        print(json.dumps(json_output))


if __name__ == "__main__":
    main()



from setuptools import setup, find_packages

setup(name='ninfo',
    version='1.0.1',
    zip_safe=False,
    description="Plugin based information gathering library",
    packages = find_packages(exclude=["tests"]),
    include_package_data=True,
    install_requires=[
        "Mako",
        "python-memcached",
        "ieeemac",
        "IPy",
    ],
    extras_require = {
        'Splunk' : ['splunk-sdk'],
    },
    entry_points = {
        'console_scripts': [
            'ninfo = ninfo:main',
        ]
    }
)
